@@ -6,7 +6,7 @@ Mojo::Redis2 - Pure-Perl non-blocking I/O Redis driver
 
 =head1 VERSION
 
-0.16
+0.17
 
 =head1 DESCRIPTION
 
@@ -131,10 +131,11 @@ use Carp ();
 use constant DEBUG => $ENV{MOJO_REDIS_DEBUG} || 0;
 use constant DEFAULT_PORT => 6379;
 
-our $VERSION = '0.16';
+our $VERSION = '0.17';
 
 my $PROTOCOL_CLASS = do {
-  my $class = $ENV{MOJO_REDIS_PROTOCOL} ||= eval "require Protocol::Redis::XS; 'Protocol::Redis::XS'" || 'Protocol::Redis';
+  my $class = $ENV{MOJO_REDIS_PROTOCOL}
+    ||= eval "require Protocol::Redis::XS; 'Protocol::Redis::XS'" || 'Protocol::Redis';
   eval "require $class; 1" or die $@;
   $class;
 };
@@ -149,7 +150,7 @@ Emitted when a new connection has been established. C<$info> is a hash ref
 with:
 
   {
-    group => $str, # basic, blpop, brpop, brpoplpush, publish, ...
+    group => $str, # basic, blocking, blpop, brpop, brpoplpush, publish, ...
     id => $connection_id,
     nb => $bool, # blocking/non-blocking
   }
@@ -191,18 +192,29 @@ Holds the encoding using for data from/to Redis. Default is UTF-8.
 
 =head2 protocol
 
-  $obj = $self->protocol;
-  $self = $self->protocol($obj);
+DEPRECATED! The protocol object cannot be shared in high load
+environments.
 
-Holds an object used to parse/generate Redis messages.
+=head2 protocol_class
+
+  $str = $self->protocol_class;
+  $self = $self->protocol_class('Protocol::Redis::XS');
+
+Holds the class name used to parse/generate Redis messages.
 Defaults to L<Protocol::Redis::XS> or L<Protocol::Redis>.
 
 L<Protocol::Redis::XS> need to be installed manually.
 
 =cut
 
-has encoding => 'UTF-8';
-has protocol => sub { $PROTOCOL_CLASS->new(api => 1); };
+has encoding       => 'UTF-8';
+has protocol_class => $PROTOCOL_CLASS;
+
+# DEPRECATED
+has protocol => sub {
+  Mojo::Util::deprecated('protocol is deprecated in favor or protocol_class');
+  $_[0]->protocol_class->new(api => 1);
+};
 
 =head2 url
 
@@ -256,6 +268,8 @@ Object constructor. Makes sure L</url> is an object.
 sub new {
   my $self = shift->SUPER::new(@_);
 
+  $self->{name} = Mojo::Util::steady_time if DEBUG;
+
   if ($self->{url} and ref $self->{url} eq '') {
     $self->{url} = "redis://$self->{url}" unless $self->{url} =~ /^redis:/;
     $self->{url} = Mojo::URL->new($self->{url});
@@ -292,8 +306,8 @@ See also L<http://redis.io/commands/brpoplpush>.
 
 =cut
 
-sub blpop { shift->_execute(blpop => BLPOP => @_); }
-sub brpop { shift->_execute(brpop => BRPOP => @_); }
+sub blpop      { shift->_execute(blpop      => BLPOP      => @_); }
+sub brpop      { shift->_execute(brpop      => BRPOP      => @_); }
 sub brpoplpush { shift->_execute(brpoplpush => BRPOPLPUSH => @_); }
 
 =head2 bulk
@@ -352,8 +366,8 @@ which will run all the Redis commands inside a transaction.
 =cut
 
 sub multi {
-  my $self = shift;
-  my @attributes = qw( encoding protocol url );
+  my $self       = shift;
+  my @attributes = qw( encoding protocol_class url );
   require Mojo::Redis2::Transaction;
   Mojo::Redis2::Transaction->new(map { $_ => $self->$_ } @attributes);
 }
@@ -396,9 +410,10 @@ See L<http://redis.io/topics/pubsub> for details.
 
 =cut
 
-sub psubscribe { shift->_pubsub(PSUBSCRIBE => @_); }
+sub psubscribe   { shift->_pubsub(PSUBSCRIBE   => @_); }
 sub punsubscribe { shift->_pubsub(PUNSUBSCRIBE => @_); }
-sub subscribe { shift->_pubsub(SUBSCRIBE => @_); }
+sub subscribe    { shift->_pubsub(SUBSCRIBE    => @_); }
+
 sub unsubscribe {
   my $self = shift;
   return $self->_pubsub(UNSUBSCRIBE => @_) if ref $_[0] eq 'ARRAY';
@@ -408,27 +423,21 @@ sub unsubscribe {
 sub DESTROY { $_[0]->{destroy} = 1; $_[0]->_cleanup; }
 
 sub _basic_operations {
-  'append',           'echo',        'decr',             'decrby',   'del',      'exists',
-  'expire',           'expireat',    'get',              'getbit',   'getrange', 'getset',
-  'hdel',             'hexists',     'hget',             'hgetall',  'hincrby',  'hkeys',
-  'hlen',             'hmget',       'hmset',            'hset',     'hsetnx',   'hvals',
-  'incr',             'incrby',      'keys',             'lindex',   'linsert',  'llen',
-  'lpop',             'lpush',       'lpushx',           'lrange',   'lrem',     'lset',
-  'ltrim',            'mget',        'move',             'mset',     'msetnx',   'persist',
-  'ping',             'publish',     'randomkey',        'rename',   'renamenx', 'rpop',
-  'rpoplpush',        'rpush',       'rpushx',           'sadd',     'scard',    'sdiff',
-  'sdiffstore',       'set',         'setbit',           'setex',    'setnx',    'setrange',
-  'sinter',           'sinterstore', 'sismember',        'smembers', 'smove',    'sort',
-  'spop',             'srandmember', 'srem',             'strlen',   'sunion',   'sunionstore',
-  'ttl',              'type',        'zadd',             'zcard',    'zcount',   'zincrby',
-  'zinterstore',      'zrange',      'zrangebyscore',    'zrank',    'zrem',     'zremrangebyrank',
-  'zremrangebyscore', 'zrevrange',   'zrevrangebyscore', 'zrevrank', 'zscore',   'zunionstore',
+  'append', 'echo', 'decr', 'decrby', 'del', 'exists', 'expire', 'expireat', 'get', 'getbit', 'getrange', 'getset',
+    'hdel', 'hexists', 'hget', 'hgetall', 'hincrby', 'hkeys', 'hlen', 'hmget', 'hmset',  'hset',   'hsetnx', 'hvals',
+    'incr', 'incrby',  'keys', 'lindex',  'linsert', 'llen',  'lpop', 'lpush', 'lpushx', 'lrange', 'lrem',   'lset',
+    'ltrim', 'mget', 'move', 'mset', 'msetnx', 'persist', 'ping', 'publish', 'randomkey', 'rename', 'renamenx', 'rpop',
+    'rpoplpush', 'rpush', 'rpushx', 'sadd', 'scard', 'sdiff', 'sdiffstore', 'set', 'setbit', 'setex', 'setnx',
+    'setrange', 'sinter', 'sinterstore', 'sismember', 'smembers', 'smove', 'sort', 'spop', 'srandmember', 'srem',
+    'strlen', 'sunion', 'sunionstore', 'ttl', 'type', 'zadd', 'zcard', 'zcount', 'zincrby', 'zinterstore', 'zrange',
+    'zrangebyscore', 'zrank', 'zrem', 'zremrangebyrank', 'zremrangebyscore', 'zrevrange', 'zrevrangebyscore',
+    'zrevrank', 'zscore', 'zunionstore',;
 }
 
-sub _blocking_group { 'blocking' }
+sub _blocking_group {'blocking'}
 
 sub _cleanup {
-  my $self = shift;
+  my $self        = shift;
   my $connections = delete $self->{connections};
 
   delete $self->{pid};
@@ -436,20 +445,21 @@ sub _cleanup {
   for my $c (values %$connections) {
     my $loop = $self->_loop($c->{nb}) or next;
     $loop->remove($c->{id}) if $c->{id};
-    $self->$_('Premature connection close', []) for grep { $_ } map { $_->[0] } @{ $c->{waiting} };
+    $self->$_('Premature connection close', []) for grep {$_} map { $_->[0] } @{$c->{waiting}};
   }
 }
 
 sub _connect {
   my ($self, $c) = @_;
-  my $url = $self->url;
-  my $db = $url->path->[0];
+  my $url      = $self->url;
+  my $db       = $url->path->[0];
   my @userinfo = split /:/, +($url->userinfo // '');
 
   Scalar::Util::weaken($self);
-  $c->{name} = $url->clone->userinfo('')->query({ g => $c->{group} })->to_string if DEBUG;
+  $c->{protocol} = $self->protocol_class->new(api => 1);
+  $c->{name} = "$self->{name}:$c->{group}:$c->{nb}" if DEBUG;
   $c->{id} = $self->_loop($c->{nb})->client(
-    { address => $url->host, port => $url->port || DEFAULT_PORT },
+    {address => $url->host, port => $url->port || DEFAULT_PORT},
     sub {
       my ($loop, $err, $stream) = @_;
 
@@ -458,7 +468,7 @@ sub _connect {
         return $self->_error($c, $err);
       }
 
-      warn "[$c->{name}] connected\n" if DEBUG;
+      warn "[$c->{name}] --- @{[$self->_debug_url($url, $c)]}\n" if DEBUG;
 
       $stream->timeout(0);
       $stream->on(close => sub { $self and $self->_error($c) });
@@ -466,10 +476,10 @@ sub _connect {
       $stream->on(read => sub { $self->_read($c, $_[1]) });
 
       # NOTE: unshift() will cause AUTH to be sent before SELECT
-      unshift @{ $c->{queue} }, [ undef, SELECT => $db ] if $db;
-      unshift @{ $c->{queue} }, [ undef, AUTH => $userinfo[1] ] if length $userinfo[1];
+      unshift @{$c->{queue}}, [undef, SELECT => $db]          if $db;
+      unshift @{$c->{queue}}, [undef, AUTH   => $userinfo[1]] if length $userinfo[1];
 
-      $self->emit(connection => { map { $_ => $c->{$_} } qw( group id nb ) });
+      $self->emit(connection => {map { $_ => $c->{$_} } qw( group id nb )});
       $self->_dequeue($c);
     },
   );
@@ -477,11 +487,24 @@ sub _connect {
   $self;
 }
 
+sub _debug_url {
+  my $self = shift;
+  my $url  = shift->clone;
+  my $c    = shift;
+
+  if (my $userinfo = $url->userinfo) {
+    $userinfo =~ s!:.*!:******!;
+    $url->userinfo($userinfo);
+  }
+
+  return $url->query({g => $c->{group}});
+}
+
 sub _dequeue {
   my ($self, $c) = @_;
-  my $loop = $self->_loop($c->{nb});
-  my $stream = $loop->stream($c->{id}) or return $self;
-  my $queue = $c->{queue};
+  my $loop   = $self->_loop($c->{nb});
+  my $stream = $loop->stream($c->{id}) or return $self;    # stream is not yet connected
+  my $queue  = $c->{queue};
   my $buf;
 
   if (!$queue->[0]) {
@@ -494,8 +517,8 @@ sub _dequeue {
     return $self;
   }
 
-  push @{ $c->{waiting} }, shift @$queue;
-  $buf = $self->_op_to_command($c->{waiting}[-1]);
+  push @{$c->{waiting}}, shift @$queue;
+  $buf = $self->_op_to_command($c);
   do { local $_ = $buf; s!\r\n!\\r\\n!g; warn "[$c->{name}] <<< ($_)\n" } if DEBUG;
   $stream->write($buf);
   $self;
@@ -508,28 +531,28 @@ sub _error {
   warn "[$c->{name}] !!! @{[$err // 'close']}\n" if DEBUG;
 
   return if $self->{destroy};
-  return $self->_connect($c) unless defined $err;
+  return $self->_requeue($c)->_connect($c) unless defined $err;
   return $self->emit(error => $err) unless @$waiting;
-  return $self->$_($err, []) for grep { $_ } map { $_->[0] } @$waiting;
+  return $self->$_($err, undef) for grep {$_} map { $_->[0] } @$waiting;
 }
 
 sub _execute {
   my $cb = ref $_[-1] eq 'CODE' ? pop : undef;
   my ($self, $group, @cmd) = @_;
 
-  $self->_cleanup unless ($self->{pid} //= $$) eq $$; # TODO: Fork safety
+  $self->_cleanup unless ($self->{pid} //= $$) eq $$;    # TODO: Fork safety
 
   if ($cb) {
-    my $c = $self->{connections}{$group} ||= { nb => 1, group => $group };
-    push @{ $c->{queue} }, [$cb, @cmd];
+    my $c = $self->{connections}{$group} ||= {nb => 1, group => $group};
+    push @{$c->{queue}}, [$cb, @cmd];
     return $self->_connect($c) unless $c->{id};
     return $self->_dequeue($c);
   }
   else {
-    my $c = $self->{connections}{$self->_blocking_group} ||= { nb => 0, group => $self->_blocking_group };
+    my $c = $self->{connections}{$self->_blocking_group} ||= {nb => 0, group => $self->_blocking_group};
     my ($err, $res);
 
-    push @{ $c->{queue} }, [sub { shift->_loop(0)->stop; ($err, $res) = @_; }, @cmd];
+    push @{$c->{queue}}, [sub { shift->_loop(0)->stop; ($err, $res) = @_; }, @cmd];
     $c->{id} ? $self->_dequeue($c) : $self->_connect($c);
     $self->_loop(0)->start;
     die "[@cmd] $err" if $err;
@@ -542,7 +565,8 @@ sub _loop {
 }
 
 sub _op_to_command {
-  my ($self, $op) = @_;
+  my ($self, $c) = @_;
+  my $op = $c->{waiting}[-1];
   my ($i, @data);
 
   for my $token (@$op) {
@@ -551,11 +575,11 @@ sub _op_to_command {
     push @data, {type => '$', data => $token};
   }
 
-  $self->protocol->encode({type => '*', data => \@data});
+  $c->{protocol}->encode({type => '*', data => \@data});
 }
 
 sub _pubsub {
-  my $cb = ref $_[-1] eq 'CODE' ? pop : sub {};
+  my $cb = ref $_[-1] eq 'CODE' ? pop : sub { };
   my ($self, $op) = (shift, shift);
   my $channels = ref $_[0] eq 'ARRAY' ? shift : [];
 
@@ -570,17 +594,17 @@ sub _pubsub {
 
 sub _read {
   my ($self, $c, $buf) = @_;
-  my $protocol = $self->protocol;
+  my $protocol = $c->{protocol};
   my $event;
 
   do { local $_ = $buf; s!\r\n!\\r\\n!g; warn "[$c->{name}] >>> ($_)\n" } if DEBUG;
   $protocol->parse($buf);
 
-  MESSAGE:
+MESSAGE:
   while (my $message = $protocol->get_message) {
     my $data = $self->_reencode_message($message);
-    my $op = shift @{ $c->{waiting} || [] };
-    my $cb = $op->[0];
+    my $op   = shift @{$c->{waiting} || []};
+    my $cb   = $op->[0];
 
     if (ref $data eq 'SCALAR') {
       $self->$cb($$data, []) if $cb;
@@ -599,23 +623,30 @@ sub _read {
 
 sub _reencode_message {
   my ($self, $message) = @_;
-  my ($type, $data) = @{$message}{qw( type data )};
+  my ($type, $data)    = @{$message}{qw( type data )};
 
   if ($type ne '*' and $self->encoding and $data) {
     $data = Encode::decode($self->encoding, $data);
   }
 
   if ($type eq '-') {
-    return \ $data;
+    return \$data;
   }
   elsif ($type ne '*') {
     return $data;
   }
   else {
-    return [ map { $self->_reencode_message($_); } @$data ];
+    return [map { $self->_reencode_message($_); } @$data];
   }
 }
 
+sub _requeue {
+  my ($self, $c) = @_;
+
+  unshift @{$c->{queue}}, grep { $_->[0] } @{delete $c->{waiting} || []};
+  return $self;
+}
+
 for my $method (__PACKAGE__->_basic_operations) {
   my $op = uc $method;
   eval "sub $method { shift->_execute(basic => $op => \@_); }; 1" or die $@;
